from abc import abstractmethod
from enum import Enum
from typing import Any, TypeVar

from typing_extensions import Protocol

from orderings.typing import Ordered, is_instance

__all__ = ("Compare", "Ordering")


class Ordering(Enum):
    """Represents ordering."""

    LESS = -1
    """The left item is *less* than the right item."""
    EQUAL = 0
    """The left item is *equal* to the right item."""
    GREATER = 1
    """The left item is *greater* than the right item."""

    def is_less(self) -> bool:
        """Checks if the ordering is [`LESS`][orderings.core.Ordering.LESS].

        Returns:
            Whether the ordering is [`LESS`][orderings.core.Ordering.LESS].
        """
        return self is type(self).LESS

    def is_equal(self) -> bool:
        """Checks if the ordering is [`EQUAL`][orderings.core.Ordering.EQUAL].

        Returns:
            Whether the ordering is [`EQUAL`][orderings.core.Ordering.EQUAL].
        """
        return self is type(self).EQUAL

    def is_greater(self) -> bool:
        """Checks if the ordering is [`GREATER`][orderings.core.Ordering.GREATER].

        Returns:
            Whether the ordering is [`GREATER`][orderings.core.Ordering.GREATER].
        """
        return self is type(self).GREATER

    def is_less_or_equal(self) -> bool:
        """Checks if the ordering is [`LESS`][orderings.core.Ordering.LESS] or
        [`EQUAL`][orderings.core.Ordering.EQUAL].
        This is equivalent to:

        ```python
        ordering.is_less() or ordering.is_equal()
        ```

        Returns:
            Whether the ordering is [`LESS`][orderings.core.Ordering.LESS]
                or [`EQUAL`][orderings.core.Ordering.EQUAL].
        """
        return self.is_less() or self.is_equal()

    def is_not_equal(self) -> bool:
        """Checks if the ordering is not [`EQUAL`][orderings.core.Ordering.EQUAL].
        This is equivalent to:

        ```python
        not ordering.is_equal()
        ```

        Returns:
            Whether the ordering is not [`EQUAL`][orderings.core.Ordering.EQUAL].
        """
        return not self.is_equal()

    def is_greater_or_equal(self) -> bool:
        """Checks if the ordering is [`GREATER`][orderings.core.Ordering.GREATER] or
        [`EQUAL`][orderings.core.Ordering.EQUAL].
        This is equivalent to:

        ```python
        ordering.is_greater() or ordering.is_equal()
        ```

        Returns:
            Whether the ordering is [`GREATER`][orderings.core.Ordering.GREATER]
                or [`EQUAL`][orderings.core.Ordering.EQUAL].
        """
        return self.is_greater() or self.is_equal()


C = TypeVar("C", bound="Compare")


class Compare(Ordered, Protocol):
    """Implements total ordering via delegation to the
    [`compare`][orderings.core.Compare.compare] method.
    """

    @abstractmethod
    def compare(self: C, other: C) -> Ordering:
        """Defines ordering via the [`Ordering`][orderings.core.Ordering] enumeration."""
        ...

    def __lt__(self: C, other: C) -> bool:
        return self.compare(other).is_less()

    def __gt__(self: C, other: C) -> bool:
        return self.compare(other).is_greater()

    def __le__(self: C, other: C) -> bool:
        return self.compare(other).is_less_or_equal()

    def __ge__(self: C, other: C) -> bool:
        return self.compare(other).is_greater_or_equal()

    def __eq__(self, other: Any) -> bool:
        return is_instance(other, type(self)) and self.compare(other).is_equal()

    def __ne__(self, other: Any) -> bool:
        return not is_instance(other, type(self)) or self.compare(other).is_not_equal()



"""Ordering enumeration and protocols."""

__description__ = "Ordering enumeration and protocols."
__url__ = "https://github.com/nekitdev/orderings"

__title__ = "orderings"
__author__ = "nekitdev"
__license__ = "MIT"
__version__ = "1.0.0"

from orderings.core import Compare, Ordering
from orderings.typing import (
    Greater, GreaterOrEqual, LenientOrdered, Less, LessOrEqual, Ordered, StrictOrdered
)

__all__ = (
    "Compare",
    "Ordering",
    "Less",
    "Greater",
    "StrictOrdered",
    "LessOrEqual",
    "GreaterOrEqual",
    "LenientOrdered",
    "Ordered",
)



from abc import abstractmethod
from builtins import isinstance as is_instance
from typing import TypeVar

from typing_extensions import Protocol

__all__ = (
    "Less",
    "Greater",
    "StrictOrdered",
    "LessOrEqual",
    "GreaterOrEqual",
    "LenientOrdered",
    "Ordered",
    "is_instance",
)

L = TypeVar("L", bound="Less")


class Less(Protocol):
    """Represents types that implement the *less* operation (`self < other`)
    where `other` is of type `Self`.
    """

    @abstractmethod
    def __lt__(self: L, other: L) -> bool:
        ...


G = TypeVar("G", bound="Greater")


class Greater(Protocol):
    """Represents types that implement the *greater* operation (`self > other`)
    where `other` is of type `Self`.
    """

    @abstractmethod
    def __gt__(self: G, other: G) -> bool:
        ...


class StrictOrdered(Less, Greater, Protocol):
    """Represents types that implement both [`Less`][orderings.typing.Less]
    and [`Greater`][orderings.typing.Greater] protocols.
    """


LE = TypeVar("LE", bound="LessOrEqual")


class LessOrEqual(Protocol):
    """Represents types that implement the *less-or-equal* operation (`self <= other`)
    where `other` is of type `Self`.
    """

    @abstractmethod
    def __le__(self: LE, other: LE) -> bool:
        ...


GE = TypeVar("GE", bound="GreaterOrEqual")


class GreaterOrEqual(Protocol):
    """Represents types that implement the *greater-or-equal* operation (`self >= other`)
    where `other` is of type `Self`.
    """

    @abstractmethod
    def __ge__(self: GE, other: GE) -> bool:
        ...


class LenientOrdered(LessOrEqual, GreaterOrEqual, Protocol):
    """Represents types that implement both [`LessOrEqual`][orderings.typing.LessOrEqual]
    and [`GreaterOrEqual`][orderings.typing.GreaterOrEqual] protocols.
    """


class Ordered(StrictOrdered, LenientOrdered, Protocol):
    """Represents types that implement both [`StrictOrdered`][orderings.typing.StrictOrdered]
    and [`LenientOrdered`][orderings.typing.LenientOrdered] protocols.
    """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['orderings']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'orderings',
    'version': '1.0.0',
    'description': 'Ordering enumeration and protocols.',
    'long_description': '# `orderings`\n\n[![License][License Badge]][License]\n[![Version][Version Badge]][Package]\n[![Downloads][Downloads Badge]][Package]\n[![Discord][Discord Badge]][Discord]\n\n[![Documentation][Documentation Badge]][Documentation]\n[![Check][Check Badge]][Actions]\n[![Test][Test Badge]][Actions]\n[![Coverage][Coverage Badge]][Coverage]\n\n> *Ordering enumeration and protocols.*\n\n## Installing\n\n**Python 3.7 or above is required.**\n\n### pip\n\nInstalling the library with `pip` is quite simple:\n\n```console\n$ pip install orderings\n```\n\nAlternatively, the library can be installed from source:\n\n```console\n$ git clone https://github.com/nekitdev/orderings.git\n$ cd orderings\n$ python -m pip install .\n```\n\n### poetry\n\nYou can add `orderings` as a dependency with the following command:\n\n```console\n$ poetry add orderings\n```\n\nOr by directly specifying it in the configuration like so:\n\n```toml\n[tool.poetry.dependencies]\norderings = "^1.0.0"\n```\n\nAlternatively, you can add it directly from the source:\n\n```toml\n[tool.poetry.dependencies.orderings]\ngit = "https://github.com/nekitdev/orderings.git"\n```\n\n## Examples\n\n### Core\n\nThe core of `orderings` is the [`Ordering`][orderings.core.Ordering] enumeration\nand the [`Compare`][orderings.core.Compare] protocol:\n\n```python\nfrom attrs import frozen\nfrom orderings import Compare, Ordering\n\nI = TypeVar("I", bound="Int")\n\n\n@frozen()\nclass Int(Compare):\n    value: int\n\n    def compare(self: I, other: I) -> Ordering:\n        self_value = self.value\n        other_value = other.value\n\n        if self_value < other_value:\n            return Ordering.LESS\n\n        if self_value > other_value:\n            return Ordering.GREATER\n\n        return Ordering.EQUAL\n```\n\n[`Compare`][orderings.core.Compare] implements all ordering operations\n(`==`, `!=`, `<`, `>`, `<=`, `>=`) using the [`compare`][orderings.core.Compare.compare] method.\n\n## Documentation\n\nYou can find the documentation [here][Documentation].\n\n## Support\n\nIf you need support with the library, you can send an [email][Email]\nor refer to the official [Discord server][Discord].\n\n## Changelog\n\nYou can find the changelog [here][Changelog].\n\n## Security Policy\n\nYou can find the Security Policy of `orderings` [here][Security].\n\n## Contributing\n\nIf you are interested in contributing to `orderings`, make sure to take a look at the\n[Contributing Guide][Contributing Guide], as well as the [Code of Conduct][Code of Conduct].\n\n## License\n\n`orderings` is licensed under the MIT License terms. See [License][License] for details.\n\n[Email]: mailto:support@nekit.dev\n\n[Discord]: https://nekit.dev/discord\n\n[Actions]: https://github.com/nekitdev/orderings/actions\n\n[Changelog]: https://github.com/nekitdev/orderings/blob/main/CHANGELOG.md\n[Code of Conduct]: https://github.com/nekitdev/orderings/blob/main/CODE_OF_CONDUCT.md\n[Contributing Guide]: https://github.com/nekitdev/orderings/blob/main/CONTRIBUTING.md\n[Security]: https://github.com/nekitdev/orderings/blob/main/SECURITY.md\n\n[License]: https://github.com/nekitdev/orderings/blob/main/LICENSE\n\n[Package]: https://pypi.org/project/orderings\n[Coverage]: https://codecov.io/gh/nekitdev/orderings\n[Documentation]: https://nekitdev.github.io/orderings\n\n[Discord Badge]: https://img.shields.io/badge/chat-discord-5865f2\n[License Badge]: https://img.shields.io/pypi/l/orderings\n[Version Badge]: https://img.shields.io/pypi/v/orderings\n[Downloads Badge]: https://img.shields.io/pypi/dm/orderings\n\n[Documentation Badge]: https://github.com/nekitdev/orderings/workflows/docs/badge.svg\n[Check Badge]: https://github.com/nekitdev/orderings/workflows/check/badge.svg\n[Test Badge]: https://github.com/nekitdev/orderings/workflows/test/badge.svg\n[Coverage Badge]: https://codecov.io/gh/nekitdev/orderings/branch/main/graph/badge.svg\n\n[orderings.core.Compare]: https://nekitdev.github.io/orderings/reference/core#orderings.core.Compare\n[orderings.core.Compare.compare]: https://nekitdev.github.io/orderings/reference/core#orderings.core.Compare.compare\n[orderings.core.Ordering]: https://nekitdev.github.io/orderings/reference/core#orderings.core.Ordering\n',
    'author': 'nekitdev',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/nekitdev/orderings',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)
